package accountCC

import (
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"encoding/json"
	"reflect"
)

type Account struct {
	DocType      AccountCCDocType `json:"doc_type,omitempty"`
	AccountID    string           `json:"account_id,omitempty"`
	WalletID     string           `json:"wallet_id,omitempty"`
	UserID       string           `json:"user_id,omitempty"`
	Type         AccountCCType    `json:"type,omitempty"`
	Status       AccountStatus    `json:"status,omitempty"`
	CurrencyType string           `json:"currency_type,omitempty"`
}

func (this Account) GetObjectType() string  {
	return "Account";
}

func (this Account) ToUserAccountInfo() *UserAccountInfo {
	return &UserAccountInfo{
		UserID: this.UserID,
		TxnAccountInfo: &TxnAccountInfo{
			AccountID: this.AccountID,
			WalletID:  this.WalletID,
		},
	}
}

func checkWalletID(this Account, manager chaincode.IManager) error {
	validObject, err := manager.GetObject(&Wallet{}, this.WalletID)
	if err != nil {
		return err
	}

	if validObject == nil {
		return errors.Errorf("Check before add. Account Id %s. Wallet %s not found", this.AccountID, this.WalletID)
	}

	wallet := validObject.(*Wallet)
	if wallet.UserID != this.UserID {
		return errors.Errorf("Check before add. Account associated with wrong UserID, expecting %s, while received %s", wallet.UserID, this.UserID)
	}

	if wallet.Status != AccountStatus_ACTIVE {
		return errors.Errorf("Check before add. Account Id %s. Wallet status not active", this.AccountID, wallet.Status)
	}

	return nil
}

func (this Account) CheckBeforeAdd(manager chaincode.IManager) error {
	// TODO when add check AccountStatus_ACTIVE : wallet.Status = model.AccountStatus_ACTIVE

	err := checkWalletID(this, manager)
	if err != nil {
		return err
	}

	return nil
}

func (this Account) CheckBeforeUpdate(updatedObject chaincode.ValidObject, manager chaincode.IManager) error {
	updatedAccount, err := Cast_Account(updatedObject)
	if err != nil {
		return err
	}

	if updatedAccount.AccountID != this.AccountID {
		return errors.Errorf("Account associated with wrong AccountID, expecting %s, while received %s", updatedAccount.AccountID, this.AccountID)
	}

	if updatedAccount.UserID != this.UserID {
		return errors.Errorf("Account associated with wrong UserID, expecting %s, while received %s", updatedAccount.UserID, this.UserID)
	}

	if updatedAccount.CurrencyType != this.CurrencyType {
		return errors.Errorf("Account associated with wrong CurrencyType, expecting %s, while received %s", updatedAccount.CurrencyType, this.CurrencyType)
	}

	err = checkWalletID(this, manager)
	if err != nil {
		return err
	}

	return nil
}

func (this Account) GetKey(manager chaincode.IManager) []string {
	return []string{this.AccountID}
}

func (this Account) New() chaincode.ValidObject {
	return &Account{}
}

func (this Account) Validate() error {
	if len(this.AccountID) == 0 {
		return errors.New("AccountID is empty")
	}
	if len(this.WalletID) == 0 {
		return errors.New("WalletID is empty")
	}
	if len(this.UserID) == 0 {
		return errors.New("UserID is empty")
	}

	if &this.Type == nil || len(string(this.Type)) == 0 {
		return errors.New("Type is empty")
	}

	if this.Type != Main && this.Type != Additional {
		return errors.New("Type need to be Main or Additional")
	}

	if &this.CurrencyType == nil || len(string(this.CurrencyType)) == 0 {
		return errors.New("CurrencyType is empty")
	}

	if &this.Status == nil || len(string(this.Status)) == 0 {
		return errors.New("Status is empty")
	}

	if this.DocType != AccountDoc {
		return errors.New("DocType not equal AccountDoc")
	}

	return nil
}

func Unmarshal_Account(data []byte) (*Account, error) {
	var obj Account
	err := json.Unmarshal(data, &obj)
	return &obj, err;
}

func Marshal_Account(data *Account) ([]byte, error) {
	return json.Marshal(data);
}

func Cast_Account(validObject chaincode.ValidObject) (*Account, error) {
	if validObject == nil {
		return nil, nil
	}

	if reflect.ValueOf(validObject).Kind() != reflect.Ptr {
		return nil, errors.New("check validObjectc is not reference")
	}

	casted, ok := validObject.(*Account)
	if !ok {
		return nil, errors.New("Problem to get object")
	}

	return casted, nil;
}
